/* Copyright 2006 by Nils Maier
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * mod_xsendfile.c: Process X-SENDFILE header cgi/scripts may set
 *     Written by Nils Maier, MaierMan@web.de, March 2006
 *
 * Whenever an X-SENDFILE header occures in the response headers drop
 * the body and send the replacement file idenfitied by this header instead.
 *
 * Method inspired by lighttpd <http://lighttpd.net/>
 * Code inspired by mod_headers, mod_rewrite and such
 *
 * Configuration:
 *   You may turn on processing in any context, where perdir config overrides server config:
 *   XSendfile On|Off - Enable/disable(default) processing
 *
 * Installation:
 *     apxs2 -cia mod_xsendfile.c
 */

/*
    version 0.8
    (peer-review required)
*/

#include "apr.h"
#include "apr_lib.h"
#include "apr_strings.h"
#include "apr_buckets.h"
#include "apr_file_io.h"

#include "apr_hash.h"
#define APR_WANT_IOVEC
#define APR_WANT_STRFUNC
#include "apr_want.h"

#include "httpd.h"
#include "http_log.h"
#include "http_config.h"
#include "http_log.h"
#define CORE_PRIVATE
#include "http_request.h"
#include "http_core.h" // needed for per-directory core-config
#include "util_filter.h"
#include "http_protocol.h" /* ap_hook_insert_error_filter */


#define AP_XSENDFILE_HEADER "X-SENDFILE"

module AP_MODULE_DECLARE_DATA xsendfile_module;

typedef enum {
    XSENDFILE_UNSET, XSENDFILE_ENABLED, XSENDFILE_DISABLED
} xsendfile_conf_active_t;

typedef struct xsendfile_conf_t
{
    xsendfile_conf_active_t enabled;
} xsendfile_conf_t;

static void *xsendfile_config_server_create(apr_pool_t *p, server_rec *s)
{
    xsendfile_conf_t *conf;

    conf = (xsendfile_conf_t *) apr_pcalloc(p, sizeof(xsendfile_conf_t));
    conf->enabled = XSENDFILE_UNSET;

    return (void*)conf;
}

static void *xsendfile_config_server_merge(apr_pool_t *p, void *basev, void *overridesv)
{
    xsendfile_conf_t *base = (xsendfile_conf_t *) basev;
    xsendfile_conf_t *overrides = (xsendfile_conf_t *) overridesv;
    xsendfile_conf_t *conf;

    conf = (xsendfile_conf_t *) apr_pcalloc(p, sizeof(xsendfile_conf_t));

    if (overrides->enabled != XSENDFILE_UNSET)
    {
        conf->enabled = overrides->enabled;
    }
    else
    {
        conf->enabled = base->enabled;
    }

    return (void*)conf;
}

static void *xsendfile_config_perdir_create(apr_pool_t *p, char *path)
{
    xsendfile_conf_t *conf;

    conf = (xsendfile_conf_t *)apr_pcalloc(p, sizeof(xsendfile_conf_t));

    conf->enabled = XSENDFILE_UNSET;

    return (void*)conf;
}

static void *xsendfile_config_perdir_merge(apr_pool_t *p, void *basev, void *overridesv)
{
    xsendfile_conf_t *base = (xsendfile_conf_t *) basev;
    xsendfile_conf_t *overrides = (xsendfile_conf_t*)overridesv;
    xsendfile_conf_t *conf;

    conf = (xsendfile_conf_t*)apr_pcalloc(p, sizeof(xsendfile_conf_t));

    if (overrides->enabled != XSENDFILE_UNSET)
    {
        conf->enabled = overrides->enabled;
    }
    else
    {
        conf->enabled = base->enabled;
    }

    return (void*)conf;
}

static const char *xsendfile_cmd_active(cmd_parms *cmd, void *perdir_confv, int flag)
{
    xsendfile_conf_t *conf = NULL;


    if (cmd->path != NULL)
    {
        conf = (xsendfile_conf_t*)perdir_confv;
    }
    else
    {
        conf = (xsendfile_conf_t*)ap_get_module_config(
            cmd->server->module_config,
            &xsendfile_module
            );
    }
    if (conf)
    {
        conf->enabled = flag ? XSENDFILE_ENABLED : XSENDFILE_DISABLED;
    }
    return NULL;
}

static void ap_xsendfile_insert_output_filter(request_rec *r)
{
    ap_add_output_filter("XSENDFILE_OUT", NULL, r, r->connection);
}

static apr_status_t ap_xsendfile_output_filter(
    ap_filter_t *f,
    apr_bucket_brigade *in
)
{
    core_dir_config *coreconf;

    xsendfile_conf_t *conf;
    xsendfile_conf_active_t enabled;

    apr_status_t status;
    apr_bucket *e = NULL;

    apr_file_t *fd = NULL;
    apr_finfo_t finfo;

    const char *file = NULL, *root = NULL;
    char *newpath = NULL;

    int errcode;

    request_rec *r = f->r;

#ifdef _DEBUG
    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, r->server, "xsendfile: output_filter");
#endif

    coreconf = (core_dir_config *)ap_get_module_config(r->per_dir_config, &core_module);
    conf = (xsendfile_conf_t *)ap_get_module_config(r->per_dir_config, &xsendfile_module);

    /*
        check if we're running after all
    */

    enabled = conf->enabled;
    if (enabled == XSENDFILE_UNSET)
    {
        conf = (xsendfile_conf_t*)ap_get_module_config(r->server->module_config, &xsendfile_module);
        enabled = conf->enabled;
    }

    /*
        should we proceed with this request?

        * sub-requests suck
        * furthermore default-handled requests suck, as they actually shouldn't be able to set headers
    */
    if (
        conf->enabled != XSENDFILE_ENABLED
        || r->status != HTTP_OK
        || r->main
        || (r->handler && strcmp(r->handler, "default-handler") == 0)
    )
    {
#ifdef _DEBUG
        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, r->server, "xsendfile: not met %d", r->status);
#endif

        ap_remove_output_filter(f);
        return ap_pass_brigade(f->next, in);
    }

    /*
        alright, look for x-sendfile
    */
    file = apr_table_get(r->headers_out, AP_XSENDFILE_HEADER);
    if (!file || !*file)
    {
#ifdef _DEBUG
        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, r->server, "xsendfile: nothing found");
#endif

        ap_remove_output_filter(f);
        return ap_pass_brigade(f->next, in);
    }

    /*
        drop *everything*
        might be pretty expensive to generate content first that goes straight to the bitbucket,
        but actually the scripts that might set this flag won't output too much anyway
    */
    while (!APR_BRIGADE_EMPTY(in))
    {
        e = APR_BRIGADE_FIRST(in);
        apr_bucket_delete(e);
    }
    r->eos_sent = 0;

    /*
        grab the current path
        somewhat nasty...
        any better ways?
    */
    root = apr_pstrdup(r->pool, r->filename);
    if ((newpath = ap_strrchr(root, '/')))
    {
        *(newpath + 1) = '\0';
    }

    /*
        alright, we now build our new path
    */
    if ((status = apr_filepath_merge(
        &newpath,
        root,
        file,
        APR_FILEPATH_SECUREROOT | APR_FILEPATH_TRUENAME,
        r->pool
    )) != OK)
    {
        ap_log_rerror(
            APLOG_MARK,
            APLOG_ERR,
            status,
            r,
            "xsendfile: unable to find file: %s",
            file
        );
        ap_die(HTTP_NOT_FOUND, r);
        return HTTP_NOT_FOUND;
    }

#ifdef _DEBUG
    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, r->server, "xsendfile: found %s", newpath);
#endif

    /*
        try open the file
    */
    if ((status = apr_file_open(
        &fd,
        newpath,
        APR_READ | APR_BINARY
#if APR_HAS_SENDFILE
        | (coreconf->enable_sendfile == ENABLE_SENDFILE_ON ?  APR_SENDFILE_ENABLED : 0)
#endif
        ,
        0,
        r->pool
    )) != APR_SUCCESS)
    {
        ap_log_rerror(
            APLOG_MARK,
            APLOG_ERR,
            status,
            r,
            "xsendfile: file permissions deny server access: %s",
            newpath
        );
        ap_die(HTTP_FORBIDDEN, r);
        return HTTP_FORBIDDEN;
    }
#if APR_HAS_SENDFILE && defined(_DEBUG)
    if (coreconf->enable_sendfile != ENABLE_SENDFILE_ON)
    {

        ap_log_error(
            APLOG_MARK,
            APLOG_WARNING,
            0,
            r->server,
            "xsendfile: sendfile configured, but not active %d",
            coreconf->enable_sendfile
            );
    }
#endif
    /*
        stat (for etag/cache/content-length stuff)
    */
    if ((status = apr_file_info_get(&finfo, APR_FINFO_NORM, fd)) != APR_SUCCESS)
    {
        ap_log_rerror(
            APLOG_MARK,
            APLOG_ERR,
            status,
            r,
            "xsendfile: unable to stat file: %s",
            newpath
        );
        apr_file_close(fd);
        ap_die(HTTP_FORBIDDEN, r);
        return HTTP_FORBIDDEN;
    }

    /*
        no inclusion of directories!
        we're serving files!
    */
    if (finfo.filetype == APR_DIR)
    {
        ap_log_rerror(
            APLOG_MARK,
            APLOG_ERR,
            APR_EBADPATH,
            r,
            "xsendfile: translated to directory %s",
            newpath
        );
        apr_file_close(fd);
        ap_die(HTTP_NOT_FOUND, r);
        return HTTP_NOT_FOUND;
    }

    /*
        need to cheat here a bit
        as etag generator will use those ;)
        and we want local_copy and cache
    */
    r->finfo.inode = finfo.inode;
    r->finfo.size = finfo.size;
    r->no_cache = r->no_local_copy = 0;
    ap_update_mtime(r, finfo.mtime);
    ap_set_last_modified(r);
    ap_set_content_length(r, finfo.size);
    ap_set_etag(r);

    /* no need anymore, we don't want people to see the origin */
    apr_table_unset(r->headers_out, AP_XSENDFILE_HEADER);
    /* as we dropped all the content this field is not valid anymore! */
    apr_table_unset(r->headers_out, "Content-Encoding");

    /* cache or something? */
    if ((errcode = ap_meets_conditions(r)) != OK)
    {
#ifdef _DEBUG
        ap_log_error(
            APLOG_MARK,
            APLOG_DEBUG,
            0,
            r->server,
            "xsendfile: met condition %d for %s",
            errcode,
            file
        );
#endif
        apr_file_close(fd);
        r->status = errcode;
    }
    else
    {
        e = apr_bucket_file_create(
            fd,
            0,
            (apr_size_t)finfo.size,
            r->pool,
            in->bucket_alloc
        );
#if APR_HAS_MMAP
        if (coreconf->enable_mmap == ENABLE_MMAP_ON)
        {
            apr_bucket_file_enable_mmap(e, 0);
        }
#if defined(_DEBUG)
        else
        {
            ap_log_error(
                APLOG_MARK,
                APLOG_WARNING,
                0,
                r->server,
                "xsendfile: mmap configured, but not active %d",
                coreconf->enable_mmap
                );
        }
#endif /* _DEBUG */
#endif /* APR_HAS_MMAP */
        APR_BRIGADE_INSERT_TAIL(in, e);
    }

    e = apr_bucket_eos_create(in->bucket_alloc);
    APR_BRIGADE_INSERT_TAIL(in, e);

    /* remove ourselves from the filter chain */
    ap_remove_output_filter(f);

#ifdef _DEBUG
    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, r->server, "xsendfile: sending %d bytes", (int)finfo.size);
#endif

    /* send the data up the stack */
    return ap_pass_brigade(f->next, in);
}

static const command_rec xsendfile_command_table[] = {
    AP_INIT_FLAG(
        "XSendFile",
        xsendfile_cmd_active,
        NULL,
        OR_FILEINFO,
        "On|Off - Enable/disable(default) processing"
        ),
    { NULL }
};

static void xsendfile_register_hooks(apr_pool_t *p)
{
    ap_hook_insert_filter(
        ap_xsendfile_insert_output_filter,
        NULL,
        NULL,
        APR_HOOK_FIRST
        );

    ap_register_output_filter(
        "XSENDFILE_OUT",
        ap_xsendfile_output_filter,
        NULL,
        AP_FTYPE_CONTENT_SET
        );
}
module AP_MODULE_DECLARE_DATA xsendfile_module =
{
    STANDARD20_MODULE_STUFF,
    xsendfile_config_perdir_create,
    xsendfile_config_perdir_merge,
    xsendfile_config_server_create,
    xsendfile_config_server_merge,
    xsendfile_command_table,
    xsendfile_register_hooks
};
